import json
from dataclasses import dataclass
from typing import KeysView

import pandas as pd
import requests

from bblocks import config
from bblocks.import_tools.common import append_new_data, ImportData


def _get_country_codes() -> None:
    """Script to fetch the country codes used by WFP. Saved as a dataframe."""

    # Get the json file from WFP website
    url: str = "https://api.hungermapdata.org/covid/data"
    file = requests.get(url).content

    # WFP codes
    wfp = json.loads(file)["countries"]

    # Create a dictionary with the iso_code: country code
    codes = {d["iso3"]: d["adm0_code"] for d in wfp}

    # Create a dataframe with the country codes and save
    df = pd.DataFrame(list(codes.items()), columns=["iso_code", "wfp_code"])
    df.to_csv(
        config.PATHS.imported_data + r"/wfp_raw/wfp_country_codes.csv", index=False
    )

    print("WFP country codes successfully downloaded.")


def _read_wfp_country_codes() -> dict:
    """Returns a dictionary with the country codes used by WFP."""

    d = pd.read_csv(config.PATHS.imported_data + r"/wfp_raw/wfp_country_codes.csv")
    return dict(zip(d["iso_code"], d["wfp_code"].astype(int)))


def _get_inflation(country_iso: str) -> None:
    """Get inflation data from VAM for a single country based on iso code"""
    url = f"https://api.vam.wfp.org/dataviz/api/GetCsv?idx=71,116&iso3={country_iso}"

    try:
        df = pd.read_csv(
            url,
            usecols=[0, 1, 2],
            skipfooter=2,
            engine="python",
            parse_dates=["Time"],
        )
    except ConnectionError:
        print(f"Data not available for {country_iso}")
        return

    df = (
        df.rename(
            columns={
                "Time": "date",
                "Value (percent)": "value",
                "Indicator": "indicator",
            }
        )
        .assign(iso_code=country_iso)
        .sort_values(by=["indicator", "date"])
    )

    df.to_csv(
        f"{config.PATHS.imported_data}/wfp_raw/{country_iso}_inflation.csv", index=False
    )


def _get_insufficient_food(code: int, iso: str) -> None:
    """Get food consumption data from WFP"""

    # API URL
    url = (
        "https://5763353767114258.eu-central-1.fc.aliyuncs.com/2016-08-15/"
        f"proxy/wfp-data-api.36/map-data/adm0/{code}/countryData.json"
    )

    # Get the json file from WFP website. If empty return None
    r = requests.get(url)

    # Check if response is invalid return None
    if r.status_code == 404:
        return None

    # Create dataframe
    data = pd.DataFrame(r.json()["fcsGraph"])

    # Code may be valid but data may be empty. If so, return None
    if len(data) == 0:
        return None

    # If data is valid, clean it and save to munged
    data = (
        data.rename(
            columns=(
                {
                    "x": "date",
                    "fcs": "value",
                    "fcsHigh": "value_high",
                    "fcsLow": "value_low",
                }
            )
        )
        .assign(date=lambda d: pd.to_datetime(d.date, format="%Y-%m-%d"), iso_code=iso)
        .pipe(append_new_data, rf"{iso}_insufficient_food.csv", "date")
    )

    data.to_csv(
        config.PATHS.imported_data + rf"/wfp_raw/{iso}_insufficient_food.csv",
        index=False,
    )


def _read_files(iso_code: str, file_name: str) -> pd.DataFrame:
    try:
        return pd.read_csv(
            config.PATHS.imported_data + rf"/wfp_raw/{iso_code}_{file_name}.csv",
            parse_dates=["date"],
        )

    except FileNotFoundError:
        return pd.DataFrame()


def _read_insufficient_food(iso_codes: list) -> pd.DataFrame:
    """Read and merge the data for the given iso codes."""

    data = pd.DataFrame()

    for iso in iso_codes:
        data = pd.concat(
            [data, _read_files(iso, "insufficient_food")], ignore_index=True
        )

    if len(data) == 0:
        print("No insufficient food data available. Run update to download data")
        return data

    return (
        data.sort_values(by=["iso_code", "date"])
        .assign(indicator="people_with_insufficient_food_consumption")
        .reset_index(drop=True)
    )


def _read_inflation(iso_codes: list) -> pd.DataFrame:
    """Read and merge the data for the given iso codes."""

    data = pd.DataFrame()

    for iso in iso_codes:
        data = pd.concat([data, _read_files(iso, "inflation")], ignore_index=True)

    if len(data) == 0:
        print("No inflation data available. Run update to download data")
        return data

    return data.sort_values(by=["iso_code", "date"]).reset_index(drop=True)


_AVAILABLE_INDICATORS: dict = {
    "inflation": _read_inflation,
    "insufficient_food": _read_insufficient_food,
}

_CODES: dict = _read_wfp_country_codes()


@dataclass
class WFPData(ImportData):
    """Class to download and read WFP inflation and insufficient food data"""

    @property
    def available_indicators(self) -> KeysView:
        """View the available indicators from WFP"""
        return _AVAILABLE_INDICATORS.keys()

    def load_indicator(self, indicator: str, **kwargs) -> None:
        """Load an indicator into the WFPData object"""
        try:
            self.indicators[indicator] = _AVAILABLE_INDICATORS[indicator](_CODES)
        except KeyError:
            raise ValueError(f"Indicator {indicator} not available")

    def update(self, **kwargs) -> None:
        """Update the data for all the indicators currently loaded"""

        if len(self.indicators) == 0:
            raise RuntimeError("No indicators loaded. Load indicators before updating")

        for indicator in self.indicators:
            if indicator == "inflation":
                _ = [_get_inflation(iso) for iso in _CODES]
            elif indicator == "insufficient_food":
                _ = [_get_insufficient_food(code, iso) for iso, code in _CODES.items()]

        print("Data correctly updated. Run `load_indicator` to load the new data")

    def get_data(
        self,
        indicators: str | list = "all",
    ) -> pd.DataFrame:
        """
        Get the data for the given indicators as a Pandas DataFrame

        Args:
            indicators: 'all', or one or more indicators.

        Returns:
            pd.DataFrame: A dataframe (long) with the selected indicators
        """

        df = pd.DataFrame()

        if indicators != "all" and isinstance(indicators, str):
            indicators = [indicators]

        if isinstance(indicators, list):
            indicators = [
                self.indicators[_] for _ in indicators if _ in list(self.indicators)
            ]
            if len(indicators) == 0:
                raise ValueError("The requested indicator has not been loaded")

        elif indicators == "all":
            indicators = self.indicators.values()

        for indicator in indicators:
            df = pd.concat([df, indicator], ignore_index=True)

        return df



import pandas as pd
import requests
from bs4 import BeautifulSoup
import os

from bblocks.config import PATHS

URL = "https://www.fao.org/worldfoodsituation/foodpricesindex/en/"


def __scrape_index_df(url: str) -> pd.DataFrame:
    """Parse FAO food price index page to retrieve data csv url"""

    headers = {"User-Agent": "Mozilla/5.0"}

    response = requests.get(url)

    if response.status_code != 200 or response.url != url:
        raise ConnectionError("FAO food price index page not found")

    soup = BeautifulSoup(response.content, "html.parser")
    href = soup.find_all(string="CSV")[0].parent.get("href")

    return pd.read_csv(
        f"https://www.fao.org/{href}",
        skiprows=2,
        storage_options=headers,
        parse_dates=["Date"],
    )


def __clean_fao_food_price_index(df: pd.DataFrame) -> pd.DataFrame:
    """Clean index dataframe"""

    return (
        df.rename(columns={"Date": "date", "Food Price Index": "food_price_index"})
        .dropna(subset="date")
        .reset_index(drop=True)
        .loc[:, lambda d: ~d.columns.str.contains("Unnamed")]
    )


def get_fao_index(
    local_path: str = f"{PATHS.imported_data}/fao_index.csv", update: bool = False
) -> pd.DataFrame:
    """Get FAO food price index

    Args:
        local_path: where the downloaded CSV will be stored

        update: if True, updates the data from the FAO website. Otherwise
            it loads the data from the local file. If a local file does not exist,
            the data will be extracted from the website.
    """

    if not os.path.exists(local_path) or update:
        df = __scrape_index_df(URL).pipe(__clean_fao_food_price_index)
        df.to_csv(local_path, index=False)

    return pd.read_csv(local_path)

get_fao_index(update=True)


""" """
import io
from zipfile import ZipFile

import pandas as pd
import requests


def read_zipped_csv(url: str, path: str) -> pd.DataFrame:
    """Reads a CSV from a zipped file on the web

    Args:
        url (str): url link to the zipped file
        path (str): path to the csv in the folder

    Returns:
        pd.DataFrame
    """

    # request url and unzip file
    try:
        response = requests.get(url)
        folder = ZipFile(io.BytesIO(response.content))

    except ConnectionError:
        raise ConnectionError("invalid url")

    # read csv
    if path not in list(folder.NameToInfo.keys()):
        raise ValueError("invalid file path")
    else:
        return pd.read_csv(folder.open(path), low_memory=False)



from typing import Optional
import pandas as pd
from operator import xor
from bblocks.cleaning_tools import clean


def __validate_cols(d: pd.DataFrame, sdate, edate, date_col, value_col, grouper):
    # If value column is None, set to 'value'
    if value_col is None:
        value_col = ["value"]

    # If value_columns is string convert to list
    if isinstance(value_col, str):
        value_col = [value_col]

    # Validate args
    if xor(sdate is not None, edate is not None):
        raise ValueError("start_date and end_date must be both specified or both None")

    for col in [date_col, *value_col]:
        if col not in d.columns:
            raise ValueError(f"{col} not found in data")

    if grouper is None:
        grouper = [c for c in d.columns if c not in value_col]

    # If group_by is a string, convert it to a list
    if isinstance(grouper, str):
        grouper = [grouper]

    # group_by cannot include the date column
    grouper = [c for c in grouper if c != date_col]

    for col in grouper:
        if col not in d.columns:
            raise ValueError(f"{col} not found in data")

    return sdate, edate, date_col, value_col, grouper


def period_avg(
    data: pd.DataFrame,
    start_date: Optional[str | int] = None,
    end_date: Optional[str | int] = None,
    date_column: str = "date",
    value_columns: str | list = None,
    group_by: Optional[str | list] = None,
) -> pd.DataFrame:
    """
    Calculate the average of (a) column(s) over a period of time.

    Args:
        data: a DataFrame with a date column (datetime or int) and one or more numeric columns
        start_date: Optionally, specify the start date of the period
        end_date: Optionally, specify the end date of the period
        date_column: the name of the date (datetime or int) column
        value_columns: one or more columns to calculate the average over
        group_by: Optionally, specify which columns to consider for the average

    Returns:
        A DataFrame with the average of the specified columns over the specified period

    """

    # Create a copy of the dataframe to avoid modifying the original
    data = data.copy(deep=True)

    # Validate args
    start_date, end_date, date_column, value_columns, group_by = __validate_cols(
        data,
        sdate=start_date,
        edate=end_date,
        date_col=date_column,
        value_col=value_columns,
        grouper=group_by,
    )

    if (start_date is not None) and (end_date is not None):
        data = data[lambda d: d[date_column].between(start_date, end_date)]

    return data.groupby(by=group_by)[value_columns].mean().reset_index()


def change_from_date(
    data: pd.DataFrame,
    date_column: str,
    start_date: str | int,
    end_date: str | int,
    value_columns: str | list = None,
    group_by: Optional[str | list] = None,
    percentage: bool = False,
) -> pd.DataFrame:
    """Calculate the change in value from a start to and end data (in #)"""

    # Create a copy of the dataframe to avoid modifying the original
    data = data.copy(deep=True)

    # Validate args
    start_date, end_date, date_column, value_columns, group_by = __validate_cols(
        data,
        sdate=start_date,
        edate=end_date,
        date_col=date_column,
        value_col=value_columns,
        grouper=group_by,
    )

    def __range_diff(s):
        return s.diff().dropna()

    def __pct_diff(s):
        return s.pct_change().dropna()

    cols = list(data.columns)

    return (
        data.loc[lambda d: d[date_column].isin([start_date, end_date])]
        .sort_values(by=[date_column] + group_by)
        .groupby(by=group_by)[value_columns]
        .apply(__range_diff if not percentage else __pct_diff)
        .reset_index()
        .filter(cols, axis=1)
    )



import os

import pandas as pd
from country_converter import convert

from bblocks.config import PATHS
from bblocks.import_tools.world_bank import WorldBankData
from bblocks.other_tools.common import Dict


def __download_income_levels():
    """Downloads fresh version of income levels from WB"""
    url = "https://databank.worldbank.org/data/download/site-content/CLASS.xlsx"

    df = pd.read_excel(
        url,
        sheet_name="List of economies",
        usecols=["Code", "Income group"],
        na_values=None,
    )

    df = df.dropna(subset=["Income group"])

    df.to_csv(PATHS.imported_data + r"/income_levels.csv", index=False)
    print("Downloaded income levels")


def __get_income_levels() -> dict:
    """Return income level dictionary"""
    file = PATHS.imported_data + r"/income_levels.csv"
    if not os.path.exists(file):
        __download_income_levels()

    return pd.read_csv(file, na_values=None, index_col="Code")["Income group"].to_dict()


__wb = WorldBankData()
(
    __wb.load_indicator("SP.DYN.LE00.IN", most_recent_only=True)
    .load_indicator("EN.POP.DNST", most_recent_only=True)
    .load_indicator("SP.POP.TOTL", most_recent_only=True)
    .load_indicator("SI.POV.DDAY", most_recent_only=True)
)


def __get_dac_codes() -> dict:
    """Return dac codes dictionary"""
    file = PATHS.imported_data + r"/oecd_codes.csv"

    return pd.read_csv(file, na_values=None, index_col="code")["iso_code"].to_dict()


def __read_flourish_geometries() -> dict:
    """Reads flourish geometries"""
    file = f"{PATHS.imported_data}/flourish_geometries.csv"
    return pd.read_csv(file, na_values=None, index_col="3-letter ISO code")[
        "geometry"
    ].to_dict()


def update_dictionaries() -> None:
    """Updates dictionaries"""
    __wb.update()
    __download_income_levels()


g20_countries: Dict = Dict(
    {
        x: convert(x, src="ISO3", to="name_short", not_found=None)
        for x in [
            "ARG",
            "AUS",
            "BRA",
            "CAN",
            "CHN",
            "FRA",
            "DEU",
            "IND",
            "IDN",
            "ITA",
            "KOR",
            "JPN",
            "MEX",
            "RUS",
            "SAU",
            "ZAF",
            "TUR",
            "GBR",
            "USA",
        ]
    }
)

eu27 = Dict(
    {
        x: convert(x, src="ISO3", to="name_short")
        for x in [
            "AUT",
            "BEL",
            "BGR",
            "HRV",
            "CZE",
            "DNK",
            "EST",
            "FIN",
            "FRA",
            "DEU",
            "GRC",
            "HUN",
            "IRL",
            "ITA",
            "LVA",
            "LTU",
            "LUX",
            "MLT",
            "NLD",
            "POL",
            "PRT",
            "ROU",
            "SVK",
            "SVN",
            "ESP",
            "SWE",
            "GBR",
        ]
    }
)

g7 = Dict(
    {
        x: convert(x, src="ISO3", to="name_short")
        for x in ["FRA", "DEU", "ITA", "GBR", "USA", "JPN", "CAN"]
    }
)

income_levels = Dict(__get_income_levels())

life_expectancy = Dict(
    __wb.get_data("SP.DYN.LE00.IN").set_index("iso_code")["value"].to_dict()
)

population_density = Dict(
    __wb.get_data("EN.POP.DNST").set_index("iso_code")["value"].to_dict()
)

population = Dict(__wb.get_data("SP.POP.TOTL").set_index("iso_code")["value"].to_dict())

dac_codes = Dict(__get_dac_codes())

flourish_geometries = Dict(__read_flourish_geometries())


from typing import Optional

import pandas as pd

from bblocks.cleaning_tools.clean import convert_id
from bblocks.dataframe_tools.common import (
    get_population_df,
    get_poverty_ratio_df,
    get_population_density_df,
    get_gdp_df,
    get_gov_expenditure_df,
)


def __validate_add_column_params(
    *,
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None,
    date_column: str | None,
) -> tuple:
    """Validate parameters to use in an *add column* function type"""

    if id_type is None:
        id_type = "regex"

    if id_column not in df.columns:
        raise ValueError(f"id_column '{id_column}' not in dataframe columns")

    if date_column is not None and date_column not in df.columns:
        raise ValueError(f"date_column '{date_column}' not in dataframe columns")

    # create id and year columns
    df["id_"] = convert_id(df[id_column], id_type)

    if date_column is not None:

        if pd.api.types.is_numeric_dtype(df[date_column]):
            try:
                df["year"] = pd.to_datetime(df[date_column], format="%Y").dt.year
            except ValueError:
                raise ValueError(
                    f"could not parse date format in '{date_column}'."
                    f"To fix, convert column to datetime"
                )
        else:
            try:
                df["year"] = pd.to_datetime(
                    df[date_column], infer_datetime_format=True
                ).dt.year

            except ValueError:
                raise ValueError(
                    f"could not parse date format in '{date_column}'."
                    f"To fix, convert column to datetime"
                )

    on_ = ["id_", "year"] if date_column is not None else ["id_"]

    return df, on_


def add_population_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    target_column: str = "population",
    update_data: bool = False,
) -> pd.DataFrame:
    """Add population column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the population
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the data isn't specified, the most recent population data from
            the world bank is used.
        target_column: the column where the population data will be stored.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the population data.
    """

    # validate parameters
    df_, on_ = __validate_add_column_params(
        df=df.copy(deep=True),
        id_column=id_column,
        id_type=id_type,
        date_column=date_column,
    )

    pop_df = get_population_df(
        most_recent_only=True if date_column is None else False,
        update_population_data=update_data,
    ).rename(columns={"iso_code": "id_"})

    df[target_column] = df_.merge(pop_df, on=on_, how="left").population

    return df


def add_poverty_ratio_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    target_column: str = "poverty_ratio",
    update_data: bool = False,
) -> pd.DataFrame:
    """Add poverty headcount column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the population
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the data isn't specified, the most recent data is used.
        target_column: the column where the population data will be stored.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the poverty data.
    """

    # validate parameters
    df_, on_ = __validate_add_column_params(
        df=df.copy(deep=True),
        id_column=id_column,
        id_type=id_type,
        date_column=date_column,
    )

    pov_df = get_poverty_ratio_df(
        most_recent_only=True if date_column is None else False,
        update_poverty_data=update_data,
    ).rename(columns={"iso_code": "id_"})

    df[target_column] = df_.merge(pov_df, on=on_, how="left").poverty_headcount_ratio

    return df


def add_population_density_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    target_column: str = "population_density",
    update_data: bool = False,
) -> pd.DataFrame:
    """Add population density column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the population
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the data isn't specified, the most recent data is used.
        target_column: the column where the population data will be stored.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the population
            density data.
    """

    # validate parameters
    df_, on_ = __validate_add_column_params(
        df=df.copy(deep=True),
        id_column=id_column,
        id_type=id_type,
        date_column=date_column,
    )

    pov_df = get_population_density_df(
        most_recent_only=True if date_column is None else False,
        update_population_data=update_data,
    ).rename(columns={"iso_code": "id_"})

    df[target_column] = df_.merge(pov_df, on=on_, how="left").population_density

    return df


def add_gdp_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    target_column: str = "gdp",
    usd: bool = True,
    include_estimates: bool = False,
    update_data: bool = False,
) -> pd.DataFrame:
    """Add GDP column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the GDP
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the date isn't specified, the most recent data is used.
        include_estimates: Whether to include years for which the WEO data is labelled as
            estimates.
        usd: Whether to add the data as US dollars or Local Currency Units.
        target_column: the column where the gdp data will be stored.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the gdp data from
            the IMF World Economic Outlook.
    """

    # validate parameters
    df_, on_ = __validate_add_column_params(
        df=df.copy(deep=True),
        id_column=id_column,
        id_type=id_type,
        date_column=date_column,
    )

    gdp_df = get_gdp_df(
        usd=usd,
        most_recent_only=True if date_column is None else False,
        include_estimates=include_estimates,
        update_gdp_data=update_data,
    ).rename(columns={"iso_code": "id_", "value": "gdp"})

    # Create a deep copy of the dataframe to avoid overwriting the original data
    _ = df.copy(deep=True).reset_index(drop=True)
    _[target_column] = df_.merge(gdp_df, on=on_, how="left").gdp

    return _


def add_gov_expenditure_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    target_column: str = "gov_exp",
    usd: bool = True,
    include_estimates: bool = False,
    update_data: bool = False,
) -> pd.DataFrame:
    """Add Government Expenditure column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the expenditure
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the date isn't specified, the most recent data is used.
        include_estimates: Whether to include years for which the WEO data is labelled as
            estimates.
        usd: Whether to add the data as US dollars or Local Currency Units.
        target_column: the column where the expenditure data will be stored.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the expenditure data from
            the IMF World Economic Outlook.
    """

    # validate parameters
    df_, on_ = __validate_add_column_params(
        df=df.copy(deep=True),
        id_column=id_column,
        id_type=id_type,
        date_column=date_column,
    )

    gov_df = get_gov_expenditure_df(
        usd=usd,
        most_recent_only=True if date_column is None else False,
        include_estimates=include_estimates,
        update_data=update_data,
    ).rename(columns={"iso_code": "id_", "value": "gov_exp"})

    # Create a deep copy of the dataframe to avoid overwriting the original data
    _ = df.copy(deep=True).reset_index(drop=True)
    _[target_column] = df_.merge(gov_df, on=on_, how="left").gov_exp

    return _


def add_gdp_share_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    value_column: str = "value",
    target_column: str = "gdp_share",
    decimals: int = 2,
    usd: bool = False,
    include_estimates: bool = False,
    update_data: bool = False,
) -> pd.DataFrame:
    """Add value as share of GDP column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the GDP
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the date isn't specified, the most recent data is used.
        value_column: the column containing the value to be converted to a share of GDP.
        decimals: the number of decimals to use in the returned column.
        include_estimates: Whether to include years for which the WEO data is labelled as
            estimates.
        usd: Whether to add the data as US dollars or Local Currency Units.
        target_column: the column where the gdp data will be stored.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the data as a share
         of gdp data, using the IMF World Economic Outlook.
    """
    kwargs = {
        k: v for k, v in dict(locals()).items() if k not in ["value_column", "decimals"]
    }

    if value_column not in df.columns:
        raise ValueError(f"value_column '{value_column}' not in dataframe columns")

    df_ = add_gdp_column(**kwargs)

    _ = df.copy(deep=True).reset_index(drop=True)

    _[target_column] = round(100 * df_[value_column] / df_[target_column], decimals)

    return _


def add_population_share_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    value_column: str = "value",
    target_column: str = "population_share",
    decimals: int = 2,
    update_data: bool = False,
) -> pd.DataFrame:
    """Add population share column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the population
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the data isn't specified, the most recent population data from
            the world bank is used.
        value_column: the column containing the value to be used in the calculation.
        target_column: the column where the population data will be stored.
        decimals: the number of decimals to use in the returned column.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing value as share of
        population.
    """

    kwargs = {
        k: v for k, v in dict(locals()).items() if k not in ["value_column", "decimals"]
    }

    if value_column not in df.columns:
        raise ValueError(f"value_column '{value_column}' not in dataframe columns")

    df_ = add_population_column(**kwargs)

    _ = df.copy(deep=True).reset_index(drop=True)

    _[target_column] = round(100 * df_[value_column] / df_[target_column], decimals)

    return _


def add_gov_exp_share_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    date_column: str | None = None,
    value_column: str = "value",
    target_column: str = "gov_exp_share",
    usd: bool = False,
    include_estimates: bool = False,
    update_data: bool = False,
) -> pd.DataFrame:
    """Add value as share of Government Expenditure column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        date_column: Optionally, a date column can be specified. If so, the expenditure data
            for that year will be used. If it's missing, it will be missing in the returned
            column as well. If the date isn't specified, the most recent data is used.
        value_column: the column containing the value to be converted to a share of expenditure.
        include_estimates: Whether to include years for which the WEO data is labelled as
            estimates.
        usd: Whether to add the data as US dollars or Local Currency Units.
        target_column: the column where the expenditure data will be stored.
        update_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the data as a share
         of expenditure, using the IMF World Economic Outlook.
    """
    kwargs = {
        k: v for k, v in dict(locals()).items() if k not in ["value_column", "decimals"]
    }

    if value_column not in df.columns:
        raise ValueError(f"value_column '{value_column}' not in dataframe columns")

    df_ = add_gov_expenditure_column(**kwargs)

    _ = df.copy(deep=True).reset_index(drop=True)

    _[target_column] = round(100 * df_[value_column] / df_[target_column], 3)

    return _


def add_income_level_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    target_column: str = "income_level",
    update_income_level_data: bool = False,
) -> pd.DataFrame:
    """Add an income levels column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DACcode, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DACcode" must be passed.
        target_column: the column where the income level data will be stored.
        update_income_level_data: whether to update the underlying data or not.

    Returns:
        DataFrame: the original DataFrame with a new column containing the income level data.
    """
    from bblocks.other_tools.dictionaries import income_levels, __download_income_levels

    # validate parameters
    df_, on_ = __validate_add_column_params(
        df=df.copy(deep=True),
        id_column=id_column,
        id_type=id_type,
        date_column=None,
    )

    if update_income_level_data:
        __download_income_levels()
        print("Downloaded income levels data")

    df[target_column] = df_["id_"].map(income_levels)

    return df


def add_short_names_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    target_column: str = "name_short",
) -> pd.DataFrame:
    """Add short names column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DAC code, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DAC" must be passed.
        target_column: the column where the short names  will be stored.

    Returns:
        DataFrame: the original DataFrame with a new column containing short names.
    """

    if id_column not in df.columns:
        raise ValueError(f"id_column '{id_column}' not in dataframe columns")

    df[target_column] = convert_id(
        df[id_column], from_type=id_type, to_type="short_name"
    )

    return df


def add_iso_codes_column(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    target_column: str = "iso_code",
) -> pd.DataFrame:
    """Add ISO3 column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DAC code, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DAC" must be passed.
        target_column: the column where the iso codes  will be stored.

    Returns:
        DataFrame: the original DataFrame with a new column containing ISO3 codes.
    """

    if id_column not in df.columns:
        raise ValueError(f"id_column '{id_column}' not in dataframe columns")

    df[target_column] = convert_id(df[id_column], from_type=id_type, to_type="ISO3")

    return df


def add_median_observation(
    df: pd.DataFrame,
    group_by: str | list = None,
    value_columns: str | list[str] = "value",
    append: bool = True,
    group_name: Optional[str] = None,
) -> pd.DataFrame:
    """Add median observation column to a dataframe

    Args:

        df: the dataframe to which the column will be added
        group_by: the column(s) by which to group the data to calculate the median.
        value_columns: the column(s) containing the values to be used for the median.
        append: if True, the median observation will be appended to the dataframe. If
            False, the median observation will be stored in a new column.
        group_name: the name of the group to be used in the id_column or as the name of
        the column containing the median observations.

    Returns:
        DataFrame: the original dataframe with added rows for the median (if append is True)
            or a new column containing the median observations (if append is False).
    """
    df_ = df.copy(deep=True)

    if isinstance(value_columns, str):
        value_columns = [value_columns]

    for c in value_columns:
        if c not in df_.columns:
            raise ValueError(f"value_column '{c}' not in dataframe columns")

    if group_by is not None:
        if isinstance(group_by, str):
            group_by = [group_by]

        for c in group_by:
            if c not in df.columns:
                raise ValueError(f"group_by column '{c}' not in dataframe columns")

    if group_by is None:
        group_by = [c for c in df_.columns if c not in value_columns]

    if group_name is None:
        group_name = "median_observation"

    median = df_.groupby(group_by)[value_columns].median()

    if append:
        _ = median.reset_index()
        _ = pd.concat([df_, _], ignore_index=True)
        _.iloc[len(df_) :] = _.iloc[len(df_) :].fillna(group_name)
        return _

    return df_.merge(
        median.reset_index(), on=group_by, how="left", suffixes=("", f" ({group_name})")
    )


def add_flourish_geometries(
    df: pd.DataFrame,
    id_column: str,
    id_type: str | None = None,
    target_column: str = "geometry",
) -> pd.DataFrame:
    """Add flourish geometries column to a dataframe

    Args:
        df: the dataframe to which the column will be added
        id_column: the column containing the name, ISO3, ISO2, DAC code, UN code, etc.
        id_type: the type of ID used in th id_column. The default 'regex' tries to infer
            using the rules from the 'country_converter' package. For the DAC codes,
            "DAC" must be passed.
        target_column: the column where the flourish geometries  will be stored.
    Returns:
        DataFrame: the original DataFrame with a new column containing the flourish geometries.
    """
    from bblocks.other_tools.dictionaries import flourish_geometries

    if id_column not in df.columns:
        raise ValueError(f"id_column '{id_column}' not in dataframe columns")

    df_, _ = __validate_add_column_params(
        df=df.copy(deep=True), id_column=id_column, id_type=id_type, date_column=None
    )

    df[target_column] = df_.id_.map(flourish_geometries)

    return df


def add_value_as_share(
    df: pd.DataFrame,
    value_col: str,
    share_of_value_col: str,
    target_col: str | None = None,
    decimals: int = 2,
) -> pd.DataFrame:

    # Copy the dataframe to avoid modifying the original one
    df = df.copy(deep=True)

    if value_col not in df.columns:
        raise ValueError(f"value_col '{value_col}' not in dataframe columns")

    if share_of_value_col not in df.columns:
        raise ValueError(
            f"share_of_value_col '{share_of_value_col}' not in dataframe columns"
        )

    if target_col is None:
        target_col = f"{value_col}_as_share_of_{share_of_value_col}"

    df[target_col] = round(100 * df[value_col] / df[share_of_value_col], decimals)

    return df
